/*
* ftcClient.cpp 
*/


#include <time.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <openssl/md5.h>
#include <ftc.h>

#include <vector>
#include <memory>
#include <iostream>
#include <fstream>
#include <iomanip>

const std::size_t max_buffer_size = 1024u*1024u*500u; // 500 MB
const std::size_t default_buffer_size = 1024u*1024u; // 1 MB
const std::size_t key_size = 16u;

double diffTime (struct timeval *t_start, struct timeval *t_finish)
{
  return t_finish->tv_sec - t_start->tv_sec + (t_finish->tv_usec - t_start->tv_usec) / 1.e6;
}

struct file_holder
{
  file_holder(FILE* f)
    : f(f) {}
  ~file_holder ()
  {
    if(f)
      fclose(f);
  }

  void close () { fclose(f); f = 0; }

  FILE* raw () { return f; }

  FILE* f;
};

bool getFile(std::string hostname, unsigned short port, std::string const& filename
             , std::size_t nbytes, std::size_t buffer_size, bool csv_output
             , std::string logFilename)
{
  struct timeval t0, t_start, t_finish;
  bool writable = false;
  unsigned long long size = 0;
  std::vector<unsigned char> md5_key(key_size);
	
  // Calculating the md5 key
  MD5(static_cast<const unsigned char*>(static_cast<const void*>(filename.c_str ()))
      ,filename.size(), &md5_key[0]);

  gettimeofday (&t0, NULL);

  file_holder fp(fopen ("/dev/null", "w"));

  if(hostname.empty())
    hostname = "127.0.0.1";

  ftc::ftc ch ( writable, hostname.c_str(), port, md5_key);

  try
  {
    ch.open();
    size = nbytes > 0 ? nbytes : ch.getSize();

    //ch.setReadBufferSize(buffer_size);

    gettimeofday (&t_start, NULL);

    ch.transferTo( 0, size, fp.raw () );
        
    gettimeofday (&t_finish, NULL);

    ch.close ();
    fp.close ();
  }
  catch (std::exception const& e)
  {
    std::cerr << "Exception: " << e.what() << std::endl;
    return false;
  }

  double responseTime = diffTime (&t0 , &t_start);
  double startTime = t_start.tv_sec + (t_start.tv_usec)/1.e6;

  double totalTime = diffTime (&t_start, &t_finish);
  
  double speed = (double)((size/1024)/totalTime);	

  std::auto_ptr<std::ofstream> logfile_stream;
  std::ostream* logfile = &std::cout; 
  if(!logFilename.empty ())
  {
    logfile_stream.reset(new std::ofstream(logFilename.c_str (), std::ios_base::app));
    logfile = logfile_stream.get ();
  }

  if(csv_output)
  {
    (*logfile) << size << "," << std::setprecision(9) << totalTime << "," << speed << "," 
               << buffer_size << "," << responseTime << "," << startTime << std::endl;
  }
  else
  {
    (*logfile) << "Arquivo: "<< filename << std::endl ;
    (*logfile) << "Tamanho: " << size << std::endl;
    logfile->precision(9);
    (*logfile) << "Tempo decorrido: " << totalTime << std::endl;
    (*logfile) << "Taxa de transferencia: " << speed  << std::endl;
  }

  return true;
}

int main (int argc, char **argv)
{
  unsigned short port = 25555;
  std::string hostname, filename, logFilename;
  size_t buffer_size = default_buffer_size; // 1MB
  size_t nbytes = 0;
  int opt;
  bool csv_output=false;

  while ((opt = getopt(argc, argv, "h:p:b:s:l:c")) != -1)
  {
    switch (opt)
    {
    case 'h':
      hostname = optarg;
      break;
    case 'l':
      logFilename = optarg;
      break;
    case 'p':
      port  = ::atoi(optarg);
      break;
    case 'c':
      csv_output=true;
      break;
    case 'b':
      buffer_size = ::atol(optarg);
      if(errno != 0) 
        buffer_size = default_buffer_size;
      else
        buffer_size = buffer_size > max_buffer_size ? max_buffer_size : buffer_size;
      break;
    case 's':
      nbytes = ::atol(optarg);
      if(errno != 0)
        nbytes = 0;
      break;
    default: /* ’?’ */
      std::cerr <<  "Usage: " << argv[0]
                << " [-h hostname] [-p port] [-b buffer_size] [-s size] [-c] filename" << std::endl;
      return EXIT_FAILURE;
    }
  }

  if (optind >= argc)
  {
    std::cout << "O nome do arquivo deve ser fornecido." <<std::endl;
    return EXIT_FAILURE;
  }

  if(!csv_output)
  {
    std::cout << "Host: " << hostname << ":" << port << std::endl;
    std::cout << "Buffer Size: " << buffer_size << std::endl;
  }

  filename = argv[optind];

  bool success = getFile(hostname, port, filename, nbytes, buffer_size
                         , csv_output, logFilename);

  if(!success)
  {
    std::cout << "Erro ao transferir o arquivo." << std::endl;
    return EXIT_FAILURE;
  }
}
