#include <ftc/DefaultDataProvider.h>
#include <ftc/FTCServerConfig.h>
#include <ftc/FTCServer.h>
#include <ftc/AccessKey.h>
#include <ftc/Log.h>

#include <iostream>
#include <string>
#include <csignal>
#include <stdio.h>
#include <openssl/md5.h>

#include "FileSystemUtils.h"

#include <boost/asio.hpp>
#include <boost/cstdint.hpp>
#include <boost/program_options.hpp>
#include <boost/filesystem.hpp>

#include <boost/version.hpp>
#if (BOOST_VERSION == 105100)
namespace boost { namespace algorithm { namespace detail {
inline unsigned hex_char_to_int ( char c );
}}}
#endif
#include <boost/algorithm/hex.hpp>

#if defined(_MSC_VER)
#pragma comment(lib, "ftc.lib")
#endif

using namespace tecgraf::ftc::v1_04_01;
using namespace std;

const size_t max_chunk_size = 1024u*1024u*500u; // 500 MB
const size_t default_chunk_size = 1024u*1024u; // 1 MB

void stop_handler(const boost::system::error_code& error, int signal_number, FTCServer* server) {
  if (error)
    return;

  cout << "Finalizando o servidor..." << endl;
  server->stop();
}

int main(int argc, char *argv[]) {
  unsigned short port = 25555;
  string hostname = "localhost", dirname;
  size_t chunk_size = default_chunk_size; // 1MB
  int opt;
  
  while ((opt = getopt(argc, argv, "h:p:c:")) != -1) {
    switch (opt) {
    case 'h':
      hostname = optarg;
      break;
    case 'p':
      port  = (unsigned short) ::atoi(optarg);
      break;
    case 'c':
      chunk_size = ::atol(optarg);
      if(errno != 0) 
        chunk_size = default_chunk_size;
      else
        chunk_size = chunk_size > max_chunk_size ? max_chunk_size : chunk_size;
      break;
    default: /* ’?’ */
      cerr <<  "Utilização: " << argv[0]
                << " [-h hostname] [-p port] [-c chunk_size] /dir/path " << endl;
      return EXIT_FAILURE;
    }
  }

  if (optind >= argc) {
    cout << "O nome do arquivo deve ser fornecido." << endl;
    return EXIT_FAILURE;
  }
  
  dirname = argv[optind];

  DefaultDataProvider dataProvider;
  FTCServerConfig config;
  config.client_timeout(10*60*1000);
  config.max_clients(1000);
  config.is_test_mode(true);
  config.port(port);
  config.hostname(hostname);
  config.chunk_size(chunk_size);
  config.log_level(tecgraf::ftc::v1_04_01::FTCServerConfig::LOG_DEBUG);
  tecgraf::ftc::v1_04_01::ftcLog.enable(config);

  FTCServer server(dataProvider, config);

  // registra para capturar o Ctrl+C ou outro sinal de término
  boost::asio::io_service io_service;
  boost::asio::signal_set stop_signals(io_service, SIGINT, SIGTERM);
#if defined(SIGQUIT)
    stop_signals.add(SIGQUIT);
#endif // defined(SIGQUIT)
  stop_signals.async_wait(boost::bind(stop_handler, _1, _2, &server));

  boost::shared_ptr<boost::thread> thread(new boost::thread(boost::bind(&FTCServer::dispatch, &server)));
  server.wait_ready(); // aguarda o servidor começar a ouvir o hostname:porta

  // Faz parse do diretório
  vector<string> files;
  getDirList(dirname,files);
  size_t size = files.size();
  unsigned char keystr[MD5_DIGEST_LENGTH];
  for(unsigned int i = 0; i < size ; i++ ) {
    std::string file = files[i];
    if( ! isDir(file) ) {
        MD5((const unsigned char*)file.c_str(), file.size(), keystr );
        AccessKey key( keystr , MD5_DIGEST_LENGTH, true );
        server.register_data_channel_info( NULL , file , &key );
        std::cout << "Servindo: " << file << std::endl;
    }
  }

  cout << "Servidor rodando em: " << server.config().hostname() << ":" << server.config().port() << endl;
  cout << "Digite Ctrl + C para terminar" << endl;  
  
  // aguarda o Ctrl+C
  io_service.run();

  // aguarda o thread do servidor finalizar
  if (thread->joinable())
    thread->join();

  cout << "Servidor finalizado." << endl;
  return 0;
}
